@@ -25,17 +25,49 @@
 #include <xf86_OSproc.h>
 #include "wcmFilter.h"
 #include <linux/serial.h>
-#include "wcmISDV4.h"
-
-#define ISDV4_QUERY "*"       /* ISDV4 query command */
-#define ISDV4_TOUCH_QUERY "%" /* ISDV4 touch query command */
-#define ISDV4_STOP "0"        /* ISDV4 stop command */
-#define ISDV4_SAMPLING "1"    /* ISDV4 sampling command */
+#include "isdv4.h"
+#include <unistd.h>
+#include <fcntl.h>
 
 #define RESET_RELATIVE(ds) do { (ds).relwheel = 0; } while (0)
 
+/* ISDV4 init process
+   This process is the same for other backends (i.e. USB).
+
+   1. isdv4Detect - called to test if device may be serial
+   2. isdv4ProbeKeys - called to fake up keybits
+   3. isdv4ParseOptions - parse ISDV4-specific options
+   4. isdv4Init - init ISDV4-specific stuff and set the tablet model.
+
+   After isdv4Init has been called, common->model points to the ISDV4 model,
+   further calls are model-specific (not that it matters for ISDV4, we only
+   have one model).
+
+   5. isdv4InitISDV4 - do whatever device-specific init is necessary
+   6. isdv4GetRanges - Query axis ranges
+
+   --- end of PreInit ---
+
+   isdv4StartTablet is called in DEVICE_ON
+   isdv4Parse is called during ReadInput.
+
+ */
+
+
+typedef struct {
+	/* Counter for dependent devices. We can only send one QUERY command to
+	   the tablet and we must not send the SAMPLING command until the last
+	   device is enabled.  */
+	int initialized_devices;
+	/* QUERY can only be run once */
+	int tablet_initialized;
+	int baudrate;
+} wcmISDV4Data;
+
 static Bool isdv4Detect(LocalDevicePtr);
+static Bool isdv4ParseOptions(LocalDevicePtr local);
 static Bool isdv4Init(LocalDevicePtr, char* id, float *version);
+static int isdv4ProbeKeys(LocalDevicePtr local);
 static void isdv4InitISDV4(WacomCommonPtr, const char* id, float version);
 static int isdv4GetRanges(LocalDevicePtr);
 static int isdv4StartTablet(LocalDevicePtr);
@@ -45,21 +77,12 @@ static int wcmSerialValidate(LocalDevicePtr local, const unsigned char* data);
 static int wcmWaitForTablet(LocalDevicePtr local, char * data, int size);
 static int wcmWriteWait(LocalDevicePtr local, const char* request);
 
-static inline int isdv4ParseQuery(const char *buffer, const size_t len,
-				  ISDV4QueryReply *reply);
-static inline int isdv4ParseTouchQuery(const char *buffer, const size_t len,
-					ISDV4TouchQueryReply *reply);
-
-static inline int isdv4ParseTouchData(const unsigned char *buffer, const size_t len,
-				      const size_t pktlen, ISDV4TouchData *touchdata);
-
-static inline int isdv4ParseCoordinateData(const unsigned char *buffer, const size_t len,
-					   ISDV4CoordinateData *coord);
-
 	WacomDeviceClass gWacomISDV4Device =
 	{
 		isdv4Detect,
+		isdv4ParseOptions,
 		isdv4Init,
+		isdv4ProbeKeys,
 	};
 
 	static WacomModel isdv4General =
@@ -72,6 +95,27 @@ static inline int isdv4ParseCoordinateData(const unsigned char *buffer, const si
 		isdv4Parse,
 	};
 
+static void memdump(LocalDevicePtr local, char *buffer, unsigned int len)
+{
+#ifdef DEBUG
+	WacomDevicePtr priv = (WacomDevicePtr)local->private;
+	WacomCommonPtr common = priv->common;
+	int i;
+
+	DBG(10, common, "memdump of ISDV4 data (len %d)\n", len);
+	/* can't use DBG macro here, need to do it manually. */
+	for (i = 0 ; i < len && common->debugLevel >= 10; i++)
+	{
+		xf86Msg(X_NONE, "%#hhx ", buffer[i]);
+		if (i % 8 == 7)
+			xf86Msg(X_NONE, "\n");
+	}
+
+	xf86Msg(X_NONE, "\n");
+#endif
+}
+
+
 static int wcmWait(int t)
 {
 	int err = xf86WaitForInput(-1, ((t) * 1000));
@@ -144,9 +188,60 @@ static int wcmSerialValidate(LocalDevicePtr local, const unsigned char* data)
 
 static Bool isdv4Detect(LocalDevicePtr local)
 {
-	WacomDevicePtr priv = (WacomDevicePtr) local->private;
+	struct serial_struct ser;
+	int rc;
+
+	rc = ioctl(local->fd, TIOCGSERIAL, &ser);
+	if (rc == -1)
+		return FALSE;
+
+	return TRUE;
+}
+
+/*****************************************************************************
+ * isdv4ParseOptions -- parse ISDV4-specific options
+ ****************************************************************************/
+static Bool isdv4ParseOptions(LocalDevicePtr local)
+{
+	WacomDevicePtr priv = (WacomDevicePtr)local->private;
 	WacomCommonPtr common = priv->common;
-	return (common->wcmForceDevice == DEVICE_ISDV4) ? 1 : 0;
+	wcmISDV4Data *isdv4data;
+	int baud;
+
+	/* Determine default baud rate */
+	baud = (common->tablet_id == 0x90)? 19200 : 38400;
+
+	baud = xf86SetIntOption(local->options, "BaudRate", baud);
+
+	switch (baud)
+	{
+		case 38400:
+		case 19200:
+			/* xf86OpenSerial() takes the baud rate from the options */
+			xf86ReplaceIntOption(local->options, "BaudRate", baud);
+			break;
+		default:
+			xf86Msg(X_ERROR, "%s: Illegal speed value "
+					"(must be 19200 or 38400).",
+					local->name);
+			return FALSE;
+	}
+
+	if (!common->private)
+	{
+		if (!(common->private = calloc(1, sizeof(wcmISDV4Data))))
+		{
+			xf86Msg(X_ERROR, "%s: failed to alloc backend-specific data.\n",
+				local->name);
+			return FALSE;
+		}
+		isdv4data = common->private;
+		isdv4data->baudrate = baud;
+		isdv4data->tablet_initialized = 0;
+		isdv4data->initialized_devices = 0;
+	}
+
+	return TRUE;
 }
 
 /*****************************************************************************
@@ -157,11 +252,12 @@ static Bool isdv4Init(LocalDevicePtr local, char* id, float *version)
 {
 	WacomDevicePtr priv = (WacomDevicePtr)local->private;
 	WacomCommonPtr common = priv->common;
+	wcmISDV4Data *isdv4data = common->private;
 
 	DBG(1, priv, "initializing ISDV4 tablet\n");
 
-	/* Initial baudrate is 38400 */
-	if (xf86SetSerialSpeed(local->fd, common->wcmISDV4Speed) < 0)
+	/* Set baudrate */
+	if (xf86SetSerialSpeed(local->fd, isdv4data->baudrate) < 0)
 		return !Success;
 
 	if(id)
@@ -172,13 +268,6 @@ static Bool isdv4Init(LocalDevicePtr local, char* id, float *version)
 	/*set the model */
 	common->wcmModel = &isdv4General;
 
-	/* Tablet PC Button is on by default */
-	common->wcmTPCButtonDefault = 1;
-
-	/* check if TPCButton was turned off by user for stylus */
-	if (priv->flags & STYLUS_ID)
-		common->wcmTPCButton = xf86SetBoolOption(local->options,
-			"TPCButton", common->wcmTPCButtonDefault);
 	return Success;
 }
 
@@ -188,8 +277,9 @@ static Bool isdv4Init(LocalDevicePtr local, char* id, float *version)
 
 static int isdv4Query(LocalDevicePtr local, const char* query, char* data)
 {
+#ifdef DEBUG
 	WacomDevicePtr priv = (WacomDevicePtr)local->private;
-	WacomCommonPtr common =	priv->common;
+#endif
 
 	DBG(1, priv, "Querying ISDV4 tablet\n");
 
@@ -202,37 +292,15 @@ static int isdv4Query(LocalDevicePtr local, const char* query, char* data)
 
 	/* Read the control data */
 	if (!wcmWaitForTablet(local, data, ISDV4_PKGLEN_TPCCTL))
-	{
-		/* Try 19200 if it is not a touch query */
-		if (common->wcmISDV4Speed != 19200 && strcmp(query, ISDV4_TOUCH_QUERY))
-		{
-			common->wcmISDV4Speed = 19200;
-			if (xf86SetSerialSpeed(local->fd, common->wcmISDV4Speed) < 0)
-				return !Success;
- 			return isdv4Query(local, query, data);
-		}
-		else
-			return !Success;
-	}
+		return !Success;
 
 	/* Control data bit check */
 	if ( !(data[0] & 0x40) )
 	{
-		/* Try 19200 if it is not a touch query */
-		if (common->wcmISDV4Speed != 19200 && strcmp(query, ISDV4_TOUCH_QUERY))
-		{
-			common->wcmISDV4Speed = 19200;
-			if (xf86SetSerialSpeed(local->fd, common->wcmISDV4Speed) < 0)
-				return !Success;
- 			return isdv4Query(local, query, data);
-		}
-		else
-		{
-			/* Reread the control data since it may fail the first time */
-			wcmWaitForTablet(local, data, ISDV4_PKGLEN_TPCCTL);
-			if ( !(data[0] & 0x40) )
-				return !Success;
-		}
+		/* Reread the control data since it may fail the first time */
+		wcmWaitForTablet(local, data, ISDV4_PKGLEN_TPCCTL);
+		if ( !(data[0] & 0x40) )
+			return !Success;
 	}
 
 	return Success;
@@ -254,9 +322,6 @@ static void isdv4InitISDV4(WacomCommonPtr common, const char* id, float version)
 	/* digitizer Y resolution in points/inch */
 	common->wcmResolY = 2540; 	
 
-	/* no touch */
-	common->tablet_id = 0x90;
-
 	/* tilt disabled */
 	common->wcmFlags &= ~TILT_ENABLED_FLAG;
 }
@@ -270,23 +335,66 @@ static int isdv4GetRanges(LocalDevicePtr local)
 	char data[BUFFER_SIZE];
 	WacomDevicePtr priv = (WacomDevicePtr)local->private;
 	WacomCommonPtr common =	priv->common;
+	wcmISDV4Data *isdv4data = common->private;
 	int ret = Success;
 
 	DBG(2, priv, "getting ISDV4 Ranges\n");
 
+	if (isdv4data->tablet_initialized)
+		goto out;
+
+	/* Set baudrate to configured value */
+	if (xf86SetSerialSpeed(local->fd, isdv4data->baudrate) < 0)
+	{
+		ret = !Success;
+		goto out;
+	}
+
 	/* Send query command to the tablet */
 	ret = isdv4Query(local, ISDV4_QUERY, data);
+	if (ret != Success)
+	{
+		int baud;
+
+		/* Try with the other baudrate */
+		baud = (isdv4data->baudrate == 38400)? 19200 : 38400;
+
+		xf86Msg(X_WARNING, "%s: Query failed with %d baud. Trying %d.\n",
+				   local->name, isdv4data->baudrate, baud);
+
+		if (xf86SetSerialSpeed(local->fd, baud) < 0)
+		{
+			ret = !Success;
+			goto out;
+		}
+
+		ret = isdv4Query(local, ISDV4_QUERY, data);
+
+		if (ret == Success) {
+			isdv4data->baudrate = baud;
+			/* xf86OpenSerial() takes the baud rate from the options */
+			xf86ReplaceIntOption(local->options, "BaudRate", baud);
+		}
+
+	}
+
 	if (ret == Success)
 	{
 		ISDV4QueryReply reply;
 		int rc;
 
-		rc = isdv4ParseQuery(data, sizeof(data), &reply);
+		rc = isdv4ParseQuery((unsigned char*)data, sizeof(data), &reply);
 		if (rc <= 0)
 		{
 			xf86Msg(X_ERROR, "%s: Error while parsing ISDV4 query.\n",
 					local->name);
-			return BadAlloc;
+			if (rc == 0)
+				DBG(2, common, "reply or len invalid.\n");
+			else
+				DBG(2, common, "header data corrupt.\n");
+			memdump(local, data, sizeof(reply));
+			ret = BadAlloc;
+			goto out;
 		}
 
 		/* transducer data */
@@ -308,23 +416,29 @@ static int isdv4GetRanges(LocalDevicePtr local)
 
 		DBG(2, priv, "Pen speed=%d "
 			"maxX=%d maxY=%d maxZ=%d resX=%d resY=%d \n",
-			common->wcmISDV4Speed, common->wcmMaxX, common->wcmMaxY,
+			isdv4data->baudrate, common->wcmMaxX, common->wcmMaxY,
 			common->wcmMaxZ, common->wcmResolX, common->wcmResolY);
 	}
 
 	/* Touch might be supported. Send a touch query command */
-	common->wcmISDV4Speed = 38400;
-	if (isdv4Query(local, ISDV4_TOUCH_QUERY, data) == Success)
+	if (isdv4data->baudrate == 38400 &&
+	    isdv4Query(local, ISDV4_TOUCH_QUERY, data) == Success)
 	{
 		ISDV4TouchQueryReply reply;
 		int rc;
 
-		rc = isdv4ParseTouchQuery(data, sizeof(data), &reply);
+		rc = isdv4ParseTouchQuery((unsigned char*)data, sizeof(data), &reply);
 		if (rc <= 0)
 		{
 			xf86Msg(X_ERROR, "%s: Error while parsing ISDV4 touch query.\n",
 					local->name);
-			return BadAlloc;
+			if (rc == 0)
+				DBG(2, common, "reply or len invalid.\n");
+			else
+				DBG(2, common, "header data corrupt.\n");
+			memdump(local, data, sizeof(reply));
+			ret = BadAlloc;
+			goto out;
 		}
 
 		switch (reply.sensor_id)
@@ -369,7 +483,7 @@ static int isdv4GetRanges(LocalDevicePtr local)
 				    xf86Msg(X_WARNING, "%s: tablet id(%x)"
 					    " mismatch with data id (0x01) \n",
 					    local->name, common->tablet_id);
-				    return ret;
+				    goto out;
 				}
 				break;
 				/* 2FGT */
@@ -380,7 +494,7 @@ static int isdv4GetRanges(LocalDevicePtr local)
 				    xf86Msg(X_WARNING, "%s: tablet id(%x)"
 					    " mismatch with data id (0x03) \n",
 					    local->name, common->tablet_id);
-				    return ret;
+				    goto out;
 				}
 				break;
 		}
@@ -403,18 +517,32 @@ static int isdv4GetRanges(LocalDevicePtr local)
 
 		DBG(2, priv, "touch speed=%d "
 			"maxTouchX=%d maxTouchY=%d TouchresX=%d TouchresY=%d \n",
-			common->wcmISDV4Speed, common->wcmMaxTouchX,
+			isdv4data->baudrate, common->wcmMaxTouchX,
 			common->wcmMaxTouchY, common->wcmTouchResolX,
 			common->wcmTouchResolY);
 	}
 
 	xf86Msg(X_INFO, "%s: serial tablet id 0x%X.\n", local->name, common->tablet_id);
 
+out:
+	if (ret == Success)
+	{
+		isdv4data->tablet_initialized = 1;
+		isdv4data->initialized_devices++;
+	}
+
 	return ret;
 }
 
 static int isdv4StartTablet(LocalDevicePtr local)
 {
+	WacomDevicePtr priv = (WacomDevicePtr)local->private;
+	WacomCommonPtr common =	priv->common;
+	wcmISDV4Data *isdv4data = common->private;
+
+	if (--isdv4data->initialized_devices)
+		return Success;
+
 	/* Tell the tablet to start sending coordinates */
 	if (!wcmWriteWait(local, ISDV4_SAMPLING))
 		return !Success;
@@ -424,6 +552,12 @@ static int isdv4StartTablet(LocalDevicePtr local)
 
 static int isdv4StopTablet(LocalDevicePtr local)
 {
+#if DEBUG
+	WacomDevicePtr priv = (WacomDevicePtr)local->private;
+	WacomCommonPtr common = priv->common;
+#endif
+	int fd_flags;
+
 	/* Send stop command to the tablet */
 	if (!wcmWriteWait(local, ISDV4_STOP))
 		return !Success;
@@ -432,6 +566,16 @@ static int isdv4StopTablet(LocalDevicePtr local)
 	if (wcmWait(250))
 		return !Success;
 
+	/* discard potential data on the line */
+	fd_flags = fcntl(local->fd, F_GETFL);
+	if (fcntl(local->fd, F_SETFL, fd_flags | O_NONBLOCK) == 0)
+	{
+		char buffer[10];
+		while (read(local->fd, buffer, sizeof(buffer)) > 0)
+			DBG(10, common, "discarding garbage data.\n");
+		fcntl(local->fd, F_SETFL, fd_flags);
+	}
+
 	return Success;
 }
 
@@ -690,7 +834,7 @@ static int wcmWaitForTablet(LocalDevicePtr local, char* answer, int size)
  * device ID. This matching only works for wacom devices (serial ID of
  * WACf), all others are simply assumed to be pen + erasor.
  */
-int isdv4ProbeKeys(LocalDevicePtr local)
+static int isdv4ProbeKeys(LocalDevicePtr local)
 {
 	int id;
 	int tablet_id = 0;
@@ -748,137 +892,15 @@ int isdv4ProbeKeys(LocalDevicePtr local)
 	 * at later stage and true knowledge of capacitive
 	 * support will be delayed until that point.
 	 */
-	if (id >= 0x0 && id <= 0x7)
-		tablet_id = 0x90;
-	else if (id >= 0x8 && id <= 0xa)
-		tablet_id = 0x93;
-	else if (id >= 0xb && id <= 0xe)
-		tablet_id = 0xe3;
-	else if (id == 0x10)
-		tablet_id = 0xe2;
+	switch(id)
+	{
+		case 0x0 ... 0x7: tablet_id = 0x90; break;
+		case 0x8 ... 0xa: tablet_id = 0x93; break;
+		case 0xb ... 0xe: tablet_id = 0xe3; break;
+		case 0x10:	  tablet_id = 0xe2; break;
+	}
 
 	return tablet_id;
 }
 
-/* Convert buffer data of buffer sized len into a query reply.
- * Returns the number of bytes read from buffer or 0 if the buffer was on
- * insufficient length. Returns -1 on parsing or internal errors.
- */
-static inline int isdv4ParseQuery(const char *buffer, const size_t len,
-				  ISDV4QueryReply *reply)
-{
-	int header, control;
-
-	if (!reply || len < ISDV4_PKGLEN_TPCCTL)
-		return 0;
-
-	header = !!(buffer[0] & HEADER_BIT);
-	control = !!(buffer[0] & CONTROL_BIT);
-
-	if (!header || !control)
-		return -1;
-
-	reply->data_id = buffer[0] & DATA_ID_MASK;
-
-	/* FIXME: big endian? */
-	reply->x_max = (buffer[1] << 9) | (buffer[2] << 2) | ((buffer[6] >> 5) & 0x3);
-	reply->y_max = (buffer[3] << 9) | (buffer[4] << 2) | ((buffer[6] >> 3) & 0x3);
-	reply->pressure_max = buffer[5] | (buffer[6] & 0x7);
-	reply->tilt_y_max = buffer[7];
-	reply->tilt_x_max = buffer[8];
-	reply->version = buffer[9] << 7 | buffer[10];
-
-	return ISDV4_PKGLEN_TPCCTL;
-}
-
-static inline int isdv4ParseTouchQuery(const char *buffer, const size_t len,
-					ISDV4TouchQueryReply *reply)
-{
-	int header, control;
-
-	if (!reply || len < ISDV4_PKGLEN_TPCCTL)
-		return 0;
-
-	header = !!(buffer[0] & HEADER_BIT);
-	control = !!(buffer[0] & CONTROL_BIT);
-
-	if (!header || !control)
-		return -1;
-
-	reply->data_id = buffer[0] & DATA_ID_MASK;
-	reply->sensor_id = buffer[2] & 0x7;
-	reply->panel_resolution = buffer[1];
-	/* FIXME: big endian? */
-	reply->x_max = (buffer[3] << 9) | (buffer[4] << 2) | ((buffer[2] >> 5) & 0x3);
-	reply->y_max = (buffer[5] << 9) | (buffer[6] << 2) | ((buffer[2] >> 3) & 0x3);
-	reply->capacity_resolution = buffer[7];
-	reply->version = buffer[9] << 7 | buffer[10];
-
-	return ISDV4_PKGLEN_TPCCTL;
-}
-
-/* pktlen defines what touch type we parse */
-static inline int isdv4ParseTouchData(const unsigned char *buffer, const size_t buff_len,
-				      const size_t pktlen, ISDV4TouchData *touchdata)
-{
-	int header, touch;
-
-	if (!touchdata || buff_len < pktlen)
-		return 0;
-
-	header = !!(buffer[0] & HEADER_BIT);
-	touch = !!(buffer[0] & TOUCH_CONTROL_BIT);
-
-	if (header != 1 || touch != 1)
-		return -1;
-
-	memset(touchdata, 0, sizeof(*touchdata));
-
-	touchdata->status = buffer[0] & 0x1;
-	/* FIXME: big endian */
-	touchdata->x = buffer[1] << 7 | buffer[2];
-	touchdata->y = buffer[3] << 7 | buffer[4];
-	if (pktlen == ISDV4_PKGLEN_TOUCH9A)
-		touchdata->capacity = buffer[5] << 7 | buffer[6];
-
-	if (pktlen == ISDV4_PKGLEN_TOUCH2FG)
-	{
-		touchdata->finger2.x = buffer[7] << 7 | buffer[8];
-		touchdata->finger2.y = buffer[9] << 7 | buffer[10];
-		touchdata->finger2.status = !!(buffer[0] & 0x2);
-		/* FIXME: is there a fg2 capacity? */
-	}
-
-	return pktlen;
-}
-
-static inline int isdv4ParseCoordinateData(const unsigned char *buffer, const size_t len,
-					   ISDV4CoordinateData *coord)
-{
-	int header, control;
-
-	if (!coord || len < ISDV4_PKGLEN_TPCPEN)
-		return 0;
-
-	header = !!(buffer[0] & HEADER_BIT);
-	control = !!(buffer[0] & TOUCH_CONTROL_BIT);
-
-	if (header != 1 || control != 0)
-		return -1;
-
-	coord->proximity = (buffer[0] >> 5) & 0x1;
-	coord->tip = buffer[0] & 0x1;
-	coord->side = (buffer[0] >> 1) & 0x1;
-	coord->eraser = (buffer[0] >> 2) & 0x1;
-	/* FIXME: big endian */
-	coord->x = (buffer[1] << 9) | (buffer[2] << 2) | ((buffer[6] >> 5) & 0x3);
-	coord->y = (buffer[3] << 9) | (buffer[4] << 2) | ((buffer[6] >> 3) & 0x3);
-
-	coord->pressure = ((buffer[6] & 0x7) << 7) | buffer[5];
-	coord->tilt_x = buffer[7];
-	coord->tilt_y = buffer[8];
-
-	return ISDV4_PKGLEN_TPCPEN;
-}
-
 /* vim: set noexpandtab shiftwidth=8: */
