@@ -26,12 +26,18 @@
 #include <asm/types.h>
 #include <linux/input.h>
 #define MAX_USB_EVENTS 32
-#define WACOM_VENDOR_ID 0x056a  /* vendor ID on the kernel device */
+
+/* vendor IDs on the kernel device */
+#define WACOM_VENDOR_ID 0x056a
+#define WALTOP_VENDOR_ID 0x172F
 
 #define DEFAULT_SUPPRESS 2      /* default suppress */
 #define MAX_SUPPRESS 100        /* max value of suppress */
 #define BUFFER_SIZE 256         /* size of reception buffer */
 #define MAXTRY 3                /* max number of try to receive magic number */
+#define MIN_ROTATION  -900      /* the minimum value of the marker pen rotation */
+#define MAX_ROTATION_RANGE 1800 /* the maximum range of the marker pen rotation */
+#define MAX_ABS_WHEEL 1023      /* the maximum value of absolute wheel */
 
 /* Default max distance to the tablet at which a proximity-out event is generated for
  * cursor device (e.g. mouse). 
@@ -110,6 +116,18 @@ struct _WacomModel
 #define ERASER_ID		0x00000008
 #define PAD_ID			0x00000010
 
+#define WCM_PEN			0x00000001
+#define WCM_TPC			0x00000002
+#define WCM_1FGT		0x00000004
+#define WCM_2FGT		0x00000008
+#define WCM_STRIP		0x00000010
+#define WCM_RING		0x00000020
+#define WCM_DUALINPUT		0x00000040
+#define WCM_ROTATION		0x00000080
+#define WCM_LCD 		0x00000100
+
+#define TabletHasFeature(common, feature) (((common)->tablet_type & (feature)) != 0)
+
 #define ABSOLUTE_FLAG		0x00000100
 #define KEEP_SHAPE_FLAG		0x00000200
 #define BAUD_19200_FLAG		0x00000400
@@ -133,8 +151,6 @@ struct _WacomModel
 					 * tablet buttons besides the strips are
 					 * treated as buttons */
 /* get/set/property */
-#if GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 3
-
 typedef struct _PROPINFO PROPINFO;
 
 struct _PROPINFO
@@ -146,7 +162,6 @@ struct _PROPINFO
 	int nSize;
 	int nDefault;
 };
-#endif /* GET_ABI_MAJOR(ABI_XINPUT_VERSION) >= 3 */
 
 struct _WacomDeviceRec
 {
@@ -231,8 +246,6 @@ struct _WacomDeviceRec
 	int tvoffsetY;		/* Y edge offset for TwinView setup */
 	int tvResolution[4];	/* twinview screens' resultion */
 	int wcmMMonitor;        /* disable/enable moving across screens in multi-monitor desktop */
-	int wcmDevOpenCount;    /* Device open count */
-	int wcmInitKeyClassCount;    /* Device InitKeyClassDeviceStruct count */
 
 	/* JEJ - throttle */
 	int throttleStart;      /* time in ticks for last wheel movement */
@@ -240,7 +253,7 @@ struct _WacomDeviceRec
 	int throttleValue;      /* current throttle value */
 
 	/* JEJ - filters */
-	int* pPressCurve;       /* pressure curve */
+	int pPressCurve[FILTER_PRESSURE_RES + 1]; /* pressure curve */
 	int nPressCtrl[4];      /* control points for curve */
 	int minPressure;	/* the minimum pressure a pen may hold */
 
@@ -323,11 +336,13 @@ struct _WacomChannel
 struct _WacomDeviceClass
 {
 	Bool (*Detect)(LocalDevicePtr local); /* detect device */
+	Bool (*ParseOptions)(LocalDevicePtr local); /* parse class-specific options */
 	Bool (*Init)(LocalDevicePtr local, char* id, float *version);   /* initialize device */
+	int  (*ProbeKeys)(LocalDevicePtr local); /* set the bits for the keys supported */
 };
 
-	extern WacomDeviceClass gWacomUSBDevice;
-	extern WacomDeviceClass gWacomISDV4Device;
+extern WacomDeviceClass gWacomUSBDevice;
+extern WacomDeviceClass gWacomISDV4Device;
 
 /******************************************************************************
  * WacomCommonRec
@@ -340,21 +355,30 @@ struct _WacomDeviceClass
  * end of record indicator or not 
 */
 #define USE_SYN_REPORTS_FLAG	8
-#define AUTODEV_FLAG		16
-
-#define DEVICE_ISDV4 		0x000C
 
 #define MAX_CHANNELS 2
 #define MAX_FINGERS  2
 
+typedef struct {
+	int wcmZoomDistance;	       /* minimum distance for a zoom touch gesture */
+	int wcmZoomDistanceDefault;    /* default minimum distance for a zoom touch gesture */
+	int wcmScrollDistance;	       /* minimum motion before sending a scroll gesture */
+	int wcmScrollDirection;	       /* store the vertical or horizontal bit in use */
+	int wcmScrollDistanceDefault;  /* default minimum motion before sending a scroll gesture */
+	int wcmGestureUsed;	       /* retain used gesture count within one in-prox event */
+	int wcmTapTime;	   	       /* minimum time between taps for a right click */
+	int wcmTapTimeDefault;         /* default minimum time between taps for a right click */
+} WacomGesturesParameters;
+
 struct _WacomCommonRec 
 {
-	/* Do not move wcmDevice, same offset as priv->name */
-	char* wcmDevice;             /* device file name */
+	/* Do not move device_path, same offset as priv->name */
+	char* device_path;           /* device file name */
 	dev_t min_maj;               /* minor/major number */
 	unsigned char wcmFlags;     /* various flags (handle tilt) */
 	int debugLevel;
 	int tablet_id;		     /* USB tablet ID */
+	int tablet_type;	     /* type (penabled/1FGT/2FGT, etc) of the tablet */
 	int fd;                      /* file descriptor to tablet */
 	int fd_refs;                 /* number of references to fd; if =0, fd is invalid */
 	unsigned long wcmKeys[NBITS(KEY_MAX)]; /* supported tool types for the device */
@@ -387,11 +411,9 @@ struct _WacomCommonRec
 	int wcmPktLength;            /* length of a packet */
 	int wcmProtocolLevel;        /* 4 for Wacom IV, 5 for Wacom V */
 	float wcmVersion;            /* ROM version */
-	int wcmForceDevice;          /* force device type (used by ISD V4) */
 	int wcmRotate;               /* rotate screen (for TabletPC) */
 	int wcmThreshold;            /* Threshold for button pressure */
 	WacomChannel wcmChannel[MAX_CHANNELS]; /* channel device state */
-	unsigned int wcmISDV4Speed;  /* serial ISDV4 link speed */
 
 	WacomDeviceClassPtr wcmDevCls; /* device class functions */
 	WacomModelPtr wcmModel;        /* model-specific functions */
@@ -402,25 +424,30 @@ struct _WacomCommonRec
 	int wcmGesture;	     	     /* disable/enable touch gesture */
 	int wcmGestureDefault;       /* default touch gesture to disable when not supported */
 	int wcmGestureMode;	       /* data is in Gesture Mode? */
+	int wcmTouchpadMode;           /* in touchpad mode? */
 	WacomDeviceState wcmGestureState[MAX_FINGERS]; /* inital state when in gesture mode */
 	int wcmCapacity;	     /* disable/enable capacity */
 	int wcmCapacityDefault;      /* default to -1 when capacity isn't supported/disabled */
 				     /* 3 when capacity is supported */
+	WacomGesturesParameters wcmGestureParameters;
 	int wcmMaxCursorDist;	     /* Max mouse distance reported so far */
 	int wcmCursorProxoutDist;    /* Max mouse distance for proxy-out max/256 units */
 	int wcmCursorProxoutDistDefault; /* Default max mouse distance for proxy-out */
 	int wcmSuppress;        	 /* transmit position on delta > supress */
 	int wcmRawSample;	     /* Number of raw data used to filter an event */
-	int wcmScaling;		     /* dealing with missing calling DevConvert case. Default 0 */
 
 	int bufpos;                        /* position with buffer */
 	unsigned char buffer[BUFFER_SIZE]; /* data read from device */
 
+	void *private;		     /* backend-specific information */
 	int wcmLastToolSerial;
 	int wcmEventCnt;
 	struct input_event wcmEvents[MAX_USB_EVENTS];  /* events for current change */
 
 	WacomToolPtr wcmTool; /* List of unique tools */
+
+	/* DO NOT TOUCH THIS. use wcmRefCommon() instead */
+	int refcnt;			/* number of devices sharing this struct */
 };
 
 #define HANDLE_TILT(comm) ((comm)->wcmFlags & TILT_ENABLED_FLAG)
@@ -457,3 +484,5 @@ struct _WacomToolArea
 };
 
 #endif /*__XF86_XF86WACOMDEFS_H */
+
+/* vim: set noexpandtab shiftwidth=8: */
