@@ -34,10 +34,13 @@
 #include <X11/Xlib.h>
 #include <X11/Xatom.h>
 #include <X11/extensions/XInput.h>
+#include <X11/XKBlib.h>
 
 #define TRACE(...) \
 	if (verbose) fprintf(stderr, "... " __VA_ARGS__)
 
+#define ArrayLength(a) (sizeof(a) / (sizeof((a)[0])))
+
 static int verbose = False;
 
 enum printformat {
@@ -52,6 +55,27 @@ enum prop_flags {
 	PROP_FLAG_WRITEONLY = 4
 };
 
+
+/**
+ * How this works:
+ * Each parameter supported by xsetwacom has a struct param_t in the global
+ * parameters[] array.
+ * For 'standard' parameters that just modify a property, the prop_* fields
+ * are set to the matching property, format and offset. The get() function
+ * then handles the retrieval  of the property, the set() function handles
+ * the modification of the property.
+ *
+ * For parameters that need more than just triggering a property, the
+ * set_func and get_func point to the matching function to modify that
+ * particular parameter. example are the ButtonX parameters that call
+ * XSetDeviceButtonMapping instead of triggering a property.
+ *
+ * device_name is filled in automatically and just used to pass around the
+ * device name (since the XDevice* doesn't store this info). printformat is
+ * a flag that changes the output required, so that the -c and -s
+ * commandline arguments work.
+ */
+
 typedef struct _param
 {
 	const char *name;	/* param name as specified by the user */
@@ -68,6 +92,7 @@ typedef struct _param
 	enum printformat printformat;
 } param_t;
 
+/* get_func/set_func calls for special parameters */
 static void map_button(Display *dpy, XDevice *dev, param_t *param, int argc, char **argv);
 static void set_mode(Display *dpy, XDevice *dev, param_t *param, int argc, char **argv);
 static void get_mode(Display *dpy, XDevice *dev, param_t *param, int argc, char **argv);
@@ -80,10 +105,6 @@ static void get_twinview(Display *dpy, XDevice *dev, param_t *param, int argc, c
 static void set_xydefault(Display *dpy, XDevice *dev, param_t *param, int argc, char **argv);
 static void get_all(Display *dpy, XDevice *dev, param_t *param, int argc, char **argv);
 static void get_param(Display *dpy, XDevice *dev, param_t *param, int argc, char **argv);
-static void not_implemented(Display *dpy, XDevice *dev, param_t *param, int argc, char **argv)
-{
-	printf("Not implemented.\n");
-}
 
 static param_t parameters[] =
 {
@@ -388,6 +409,39 @@ static param_t parameters[] =
 		.prop_offset = 0,
 		.prop_flags = PROP_FLAG_BOOLEAN
 	},
+	{
+		.name = "Gesture",
+		.desc = "Turns on/off multi-touch gesture events "
+		"(default is enable/on). ",
+		.prop_name = WACOM_PROP_ENABLE_GESTURE,
+		.prop_format = 8,
+		.prop_offset = 0,
+		.prop_flags = PROP_FLAG_BOOLEAN
+	},
+	{
+		.name = "ZoomDistance",
+		.desc = "Minimum distance for a zoom gesture "
+		"(default is 50). ",
+		.prop_name = WACOM_PROP_GESTURE_PARAMETERS,
+		.prop_format = 32,
+		.prop_offset = 0,
+	},
+	{
+		.name = "ScrollDistance",
+		.desc = "Minimum motion before sending a scroll gesture "
+		"(default is 20). ",
+		.prop_name = WACOM_PROP_GESTURE_PARAMETERS,
+		.prop_format = 32,
+		.prop_offset = 0,
+	},
+	{
+		.name = "TapTime",
+		.desc = "Minimum time between taps for a right click "
+		"(default is 250). ",
+		.prop_name = WACOM_PROP_GESTURE_PARAMETERS,
+		.prop_format = 32,
+		.prop_offset = 0,
+	},
 	{
 		.name = "Capacity",
 		.desc = "Touch sensitivity level (default is 3, "
@@ -806,18 +860,6 @@ static param_t parameters[] =
 		.prop_offset = 0,
 		.prop_flags = PROP_FLAG_READONLY
 	},
-	{
-		.name = "NumScreen",
-		.desc = "Returns number of screens configured for the desktop. ",
-		.set_func = not_implemented,
-		.get_func = not_implemented,
-	},
-	{
-		.name = "XScaling",
-		.desc = "Returns the status of XSCALING is set or not. ",
-		.set_func = not_implemented,
-		.get_func = not_implemented,
-	},
 	{
 		.name = "all",
 		.desc = "Get value for all parameters.",
@@ -827,6 +869,57 @@ static param_t parameters[] =
 	{ NULL }
 };
 
+struct modifier {
+	char *name;
+	char *converted;
+};
+
+static struct modifier modifiers[] = {
+	{"ctrl", "Control_L"},
+	{"ctl", "Control_L"},
+	{"control", "Control_L"},
+	{"lctrl", "Control_L"},
+	{"rctrl", "Control_R"},
+
+	{"meta", "Meta_L"},
+	{"lmeta", "Meta_L"},
+	{"rmeta", "Meta_R"},
+
+	{"alt", "Alt_L"},
+	{"lalt", "Alt_L"},
+	{"ralt", "Alt_R"},
+
+	{"shift", "Shift_L"},
+	{"lshift", "Shift_L"},
+	{"rshift", "Shift_R"},
+
+	{ NULL, NULL }
+};
+
+static struct modifier specialkeys[] = {
+	{"f1", "F1"}, {"f2", "F2"}, {"f3", "F3"},
+	{"f4", "F4"}, {"f5", "F5"}, {"f6", "F6"},
+	{"f7", "F7"}, {"f8", "F8"}, {"f9", "F9"},
+	{"f10", "F10"}, {"f11", "F11"}, {"f12", "F12"},
+	{"f13", "F13"}, {"f14", "F14"}, {"f15", "F15"},
+	{"f16", "F16"}, {"f17", "F17"}, {"f18", "F18"},
+	{"f19", "F19"}, {"f20", "F20"}, {"f21", "F21"},
+	{"f22", "F22"}, {"f23", "F23"}, {"f24", "F24"},
+	{"f25", "F25"}, {"f26", "F26"}, {"f27", "F27"},
+	{"f28", "F28"}, {"f29", "F29"}, {"f30", "F30"},
+	{"f31", "F31"}, {"f32", "F32"}, {"f33", "F33"},
+	{"f34", "F34"}, {"f35", "F35"},
+
+	{"esc", "Escape"}, {"Esc", "Escape"},
+
+	{"up", "Up"}, {"down", "Down"},
+	{"left", "Left"}, {"right", "Right"},
+
+	{"backspace", "BackSpace"}, {"Backspace", "BackSpace"},
+
+	{ NULL, NULL }
+};
+
 static param_t* find_parameter(char *name)
 {
 	param_t *param = NULL;
@@ -878,7 +971,7 @@ static void usage(void)
 	printf(
 	"\nCommands:\n"
 	" --list [dev|param]           - display known devices, parameters \n"
-	" --list mod                   - display supported modifier and specific keys for keystokes [not implemented}\n"
+	" --list mod                   - display supported modifier and specific keys for keystrokes\n"
 	" --set dev_name param [values...] - set device parameter by name\n"
 	" --get dev_name param [param...] - get current device parameter(s) value by name\n");
 }
@@ -945,6 +1038,32 @@ static XDevice* find_device(Display *display, char *name)
 	return dev;
 }
 
+/* Return True if the given device has the property, or False otherwise */
+static Bool test_property(Display *dpy, XDevice* dev, Atom prop)
+{
+	int nprops_return;
+	Atom *properties;
+	int found = False;
+
+	/* if no property is required, return success */
+	if (prop == None)
+		return True;
+
+	properties = XListDeviceProperties(dpy, dev, &nprops_return);
+
+	while(nprops_return--)
+	{
+		if (properties[nprops_return] == prop)
+		{
+			found = True;
+			break;
+		}
+	}
+
+	XFree(properties);
+	return found;
+}
+
 static void list_one_device(Display *dpy, XDeviceInfo *info)
 {
 	static int	wacom_prop = 0;
@@ -1029,12 +1148,25 @@ static void list_param(Display *dpy)
 
 	while(param->name)
 	{
-		printf("%-16s - %16s%s\n", param->name, param->desc,
-			(param->set_func == not_implemented) ? " [not implemented]" : "");
+		printf("%-16s - %16s\n", param->name, param->desc);
 		param++;
 	}
 }
 
+static void list_mod(Display *dpy)
+{
+	struct modifier *m = modifiers;
+
+	printf("%d modifiers are supported:\n", ArrayLength(modifiers) - 1);
+	while(m->name)
+		printf("	%s\n", m++->name);
+
+	printf("\n%d specialkeys are supported:\n", ArrayLength(specialkeys) - 1);
+	m = specialkeys;
+	while(m->name)
+		printf("	%s\n", m++->name);
+}
+
 static void list(Display *dpy, int argc, char **argv)
 {
 	TRACE("'list' requested.\n");
@@ -1044,59 +1176,29 @@ static void list(Display *dpy, int argc, char **argv)
 		list_devices(dpy);
 	else if (strcmp(argv[0], "param") == 0)
 		list_param(dpy);
+	else if (strcmp(argv[0], "mod") == 0)
+		list_mod(dpy);
 	else
 		printf("unknown argument to list.\n");
 }
-
 /*
  * Convert a list of random special keys to strings that can be passed into
  * XStringToKeysym
  */
 static char *convert_specialkey(const char *modifier)
 {
-	struct modifier {
-		char *name;
-		char *converted;
-	} modmap[] = {
-		{"ctrl", "Control_L"},
-		{"ctl", "Control_L"},
-		{"control", "Control_L"},
-		{"lctrl", "Control_L"},
-		{"rctrl", "Control_R"},
-
-		{"meta", "Meta_L"},
-		{"lmeta", "Meta_L"},
-		{"rmeta", "Meta_R"},
-
-		{"alt", "Alt_L"},
-		{"lalt", "Alt_L"},
-		{"ralt", "Alt_R"},
-
-		{"shift", "Shift_L"},
-		{"lshift", "Shift_L"},
-		{"rshift", "Shift_R"},
-
-		{"f1", "F1"}, {"f2", "F2"}, {"f3", "F3"},
-		{"f4", "F4"}, {"f5", "F5"}, {"f6", "F6"},
-		{"f7", "F7"}, {"f8", "F8"}, {"f9", "F9"},
-		{"f10", "F10"}, {"f11", "F11"}, {"f12", "F12"},
-		{"f13", "F13"}, {"f14", "F14"}, {"f15", "F15"},
-		{"f16", "F16"}, {"f17", "F17"}, {"f18", "F18"},
-		{"f19", "F19"}, {"f20", "F20"}, {"f21", "F21"},
-		{"f22", "F22"}, {"f23", "F23"}, {"f24", "F24"},
-		{"f25", "F25"}, {"f26", "F26"}, {"f27", "F27"},
-		{"f28", "F28"}, {"f29", "F29"}, {"f30", "F30"},
-		{"f31", "F31"}, {"f32", "F32"}, {"f33", "F33"},
-		{"f34", "F34"}, {"f35", "F35"},
-
-		{ NULL, NULL }
-	};
-
-	struct modifier *m = modmap;
+	struct modifier *m = modifiers;
 
 	while(m->name && strcasecmp(modifier, m->name))
 		m++;
 
+	if (!m->name)
+	{
+		m = specialkeys;
+		while(m->name && strcasecmp(modifier, m->name))
+			m++;
+	}
+
 	return m->converted ? m->converted : (char*)modifier;
 }
 
@@ -1126,19 +1228,57 @@ static int is_modifier(const char* modifier)
 	return 0;
 }
 
-static int special_map_keystrokes(int argc, char **argv, unsigned long *ndata, unsigned long* data);
-static int special_map_button(int argc, char **argv, unsigned long *ndata, unsigned long* data);
+static int special_map_keystrokes(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long* data);
+static int special_map_button(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long* data);
+static int special_map_core(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data);
+static int special_map_modetoggle(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data);
+static int special_map_displaytoggle(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data);
 
 /* Valid keywords for the --set ButtonX options */
 struct keywords {
 	const char *keyword;
-	int (*func)(int, char **, unsigned long*, unsigned long *);
+	int (*func)(Display*, int, char **, unsigned long*, unsigned long *);
 } keywords[] = {
 	{"key", special_map_keystrokes},
 	{"button", special_map_button},
+	{"core", special_map_core},
+	{"modetoggle", special_map_modetoggle},
+	{"displaytoggle", special_map_displaytoggle},
 	{ NULL, NULL }
 };
 
+/* the "core" keyword isn't supported anymore, we just have this here to
+   tell people that. */
+static int special_map_core(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data)
+{
+	static int once_only = 1;
+	if (once_only)
+	{
+		printf ("Note: The \"core\" keyword is not supported anymore and "
+			"will be ignored.\n");
+		once_only = 0;
+	}
+	return 0;
+}
+
+static int special_map_modetoggle(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data)
+{
+	data[*ndata] = AC_MODETOGGLE;
+
+	*ndata += 1;
+
+	return 0;
+}
+
+static int special_map_displaytoggle(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data)
+{
+	data[*ndata] = AC_DISPLAYTOGGLE;
+
+	*ndata += 1;
+
+	return 0;
+}
+
 static inline int is_valid_keyword(const char *keyword)
 {
 	struct keywords *kw = keywords;
@@ -1152,7 +1292,7 @@ static inline int is_valid_keyword(const char *keyword)
 	return 0;
 }
 
-static int special_map_button(int argc, char **argv, unsigned long *ndata, unsigned long *data)
+static int special_map_button(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long *data)
 {
 	int nitems = 0;
 	int i;
@@ -1168,9 +1308,6 @@ static int special_map_button(int argc, char **argv, unsigned long *ndata, unsig
 			if (is_valid_keyword(btn))
 				break;
 
-			if (sscanf(btn, "%d", &button) != 1)
-				return nitems;
-
 			switch (btn[0])
 			{
 				case '+': need_press = 1; break;
@@ -1182,6 +1319,10 @@ static int special_map_button(int argc, char **argv, unsigned long *ndata, unsig
 		} else
 			need_press = need_release = 1;
 
+		if (sscanf(btn, "%d", &button) != 1)
+			return nitems;
+
+
 		TRACE("Button map %d [%s,%s]\n", abs(button),
 				need_press ?  "press" : "",
 				need_release ?  "release" : "");
@@ -1196,11 +1337,44 @@ static int special_map_button(int argc, char **argv, unsigned long *ndata, unsig
 	return nitems;
 }
 
+/* Return the first keycode to have the required keysym in the current group.
+   TODOs:
+   - parse other groups as well (do we need this?)
+   - for keysyms not on level 0, return the keycodes for the modifiers as
+     well
+*/
+static int keysym_to_keycode(Display *dpy, KeySym sym)
+{
+	static XkbDescPtr xkb = NULL;
+	XkbStateRec state;
+	int group;
+	int kc = 0;
+
+
+	if (!xkb)
+		xkb = XkbGetKeyboard(dpy, XkbAllComponentsMask, XkbUseCoreKbd);
+	XkbGetState(dpy, XkbUseCoreKbd, &state);
+	group = state.group;
+
+	for (kc = xkb->min_key_code; kc <= xkb->max_key_code; kc++)
+	{
+		KeySym* ks;
+		int i;
+
+		ks = XkbKeySymsPtr(xkb, kc);
+		for (i = 0; i < XkbKeyGroupWidth(xkb, kc, state.group); i++)
+			if (ks[i] == sym)
+				goto out;
+	}
+
+out:
+	return kc;
+}
 /*
    Map gibberish like "ctrl alt f2" into the matching AC_KEY values.
    Returns 1 on success or 0 otherwise.
  */
-static int special_map_keystrokes(int argc, char **argv, unsigned long *ndata, unsigned long* data)
+static int special_map_keystrokes(Display *dpy, int argc, char **argv, unsigned long *ndata, unsigned long* data)
 {
 	int i;
 	int nitems = 0;
@@ -1208,6 +1382,7 @@ static int special_map_keystrokes(int argc, char **argv, unsigned long *ndata, u
 	for (i = 0; i < argc; i++)
 	{
 		KeySym ks;
+		KeyCode kc;
 		int need_press = 0, need_release = 0;
 		char *key = argv[i];
 
@@ -1248,12 +1423,14 @@ static int special_map_keystrokes(int argc, char **argv, unsigned long *ndata, u
 			need_press = need_release = 1;
 
 		ks = XStringToKeysym(key);
-		if (need_press)
-			data[*ndata + nitems++] = AC_KEY | AC_KEYBTNPRESS | ks;
-		if (need_release)
-			data[*ndata + nitems++] = AC_KEY | ks;
+		kc = keysym_to_keycode(dpy, ks);
 
-		TRACE("Key map %ld ('%s') [%s,%s]\n", ks,
+		if (need_press)
+			data[*ndata + nitems++] = AC_KEY | AC_KEYBTNPRESS | kc;
+		if (need_release)
+			data[*ndata + nitems++] = AC_KEY | kc;
+
+		TRACE("Key map %ld (%d, '%s') [%s,%s]\n", ks, kc,
 				XKeysymToString(ks),
 				need_press ?  "press" : "",
 				need_release ?  "release" : "");
@@ -1371,7 +1548,7 @@ static void special_map_buttons(Display *dpy, XDevice *dev, param_t* param, int
 			int parsed = 0;
 			if (strcasecmp(words[i], keywords[j].keyword) == 0)
 			{
-				parsed = keywords[j].func(nwords - i - 1,
+				parsed = keywords[j].func(dpy, nwords - i - 1,
 							  &words[i + 1],
 							  &nitems, data);
 				i += parsed;
@@ -1525,8 +1702,12 @@ static void set_rotate(Display *dpy, XDevice *dev, param_t* param, int argc, cha
 		rotation = 3;
 	else if (strcasecmp(argv[0], "NONE") == 0)
 		rotation = 0;
-	else
-		goto error;
+	else if (strlen(argv[0]) == 1)
+	{
+		rotation = atoi(argv[0]);
+		if (rotation < 0 || rotation > 3)
+			goto error;
+	}
 
 	prop = XInternAtom(dpy, param->prop_name, True);
 	if (!prop)
@@ -1628,7 +1809,7 @@ static void set(Display *dpy, int argc, char **argv)
 {
 	param_t *param;
 	XDevice *dev = NULL;
-	Atom prop, type;
+	Atom prop = None, type;
 	int format;
 	unsigned char* data = NULL;
 	unsigned long nitems, bytes_after;
@@ -1663,17 +1844,22 @@ static void set(Display *dpy, int argc, char **argv)
 	{
 		printf("'%s' is a read-only option.\n", argv[1]);
 		goto out;
-	} else if (param->set_func)
-	{
-		param->set_func(dpy, dev, param, argc - 2, &argv[2]);
-		goto out;
 	}
 
-	prop = XInternAtom(dpy, param->prop_name, True);
-	if (!prop)
+	if (param->prop_name)
 	{
-		fprintf(stderr, "Property for '%s' not available.\n",
-			param->name);
+		prop = XInternAtom(dpy, param->prop_name, True);
+		if (!prop || !test_property(dpy, dev, prop))
+		{
+			printf("Property '%s' does not exist on device.\n",
+				param->prop_name);
+			goto out;
+		}
+	}
+
+	if (param->set_func)
+	{
+		param->set_func(dpy, dev, param, argc - 2, &argv[2]);
 		goto out;
 	}
 
@@ -1971,25 +2157,28 @@ static void get(Display *dpy, enum printformat printformat, int argc, char **arg
 
 static void get_param(Display *dpy, XDevice *dev, param_t *param, int argc, char **argv)
 {
-	Atom prop, type;
+	Atom prop = None, type;
 	int format;
 	unsigned char* data;
 	unsigned long nitems, bytes_after;
 
+	if (param->prop_name)
+	{
+		prop = XInternAtom(dpy, param->prop_name, True);
+		if (!prop || !test_property(dpy, dev, prop))
+		{
+			printf("Property '%s' does not exist on device.\n",
+				param->prop_name);
+			return;
+		}
+	}
+
 	if (param->get_func)
 	{
 		param->get_func(dpy, dev, param, argc, argv);
 		return;
 	}
 
-	prop = XInternAtom(dpy, param->prop_name, True);
-	if (!prop)
-	{
-		fprintf(stderr, "Property for '%s' not available.\n",
-			param->name);
-		return;
-	}
-
 	XGetDeviceProperty(dpy, dev, prop, 0, 1000, False, AnyPropertyType,
 				&type, &format, &nitems, &bytes_after, &data);
 
@@ -2026,7 +2215,7 @@ static void get_param(Display *dpy, XDevice *dev, param_t *param, int argc, char
 
 int main (int argc, char **argv)
 {
-	char c;
+	int c;
 	int optidx;
 	char *display = NULL;
 	Display *dpy;
